                                      #----------------------------#
                                      #       FORESTAPI.py         #
                                      #   Samples for FOREST APIs  #
                                      #                            #
                                      #                   16.06.21 #
                                      #                            #
                                      #            Olivier Casile  #
                                      #  forest_othello@orange.fr  #
                                      #----------------------------#


import sys
from othpy import *
from forestpy import *

# Preliminary notes:
# - All the Forest APIs have French names, sorry for the inconvenience.
#   This is for historical reason and changing it would impact the massive amount of Python code surrounding the Forest program.
# - Although the FOREST algorithms are quite robust, there is no checking of the validity of the parameters you provide to the API.
#   Entering wrong parameters will deliver unpredictable result, or simply crash.


# Linear Positions of game squares, in WTHOR game format
#       A  B  C  D  E  F  G  H
#   1  11 12 13 14 15 16 17 18
#   2  21 22 23 24 25 26 27 28
#   3  31 32 33 34 35 36 37 38
#   4  41 42 43 44 45 46 47 48
#   5  51 52 53 54 55 56 57 58
#   6  61 62 63 64 65 66 67 68
#   7  71 72 73 74 75 76 77 78
#   8  81 82 83 84 85 86 87 88
F5 = 56
E6 = 65
C4 = 43
D3 = 34

# Representation des cases du Jeu dans les configs
EMPTY = 0             # Empty Square
WHITE = 1             # White Disc (O)
BLACK = 2             # Black Disk (X)
EDGE  = 4             # Edge of the Board

# Initial Board Game
E = EDGE
e = EMPTY
O = WHITE
X = BLACK
INIT_GAME = bytes([
    E, E, E, E, E, E, E, E, E, E,
    E, e, e, e, e, e, e, e, e, E,
    E, e, e, e, e, e, e, e, e, E,
    E, e, e, e, e, e, e, e, e, E,
    E, e, e, e, O, X, e, e, e, E,
    E, e, e, e, X, O, e, e, e, E,
    E, e, e, e, e, e, e, e, e, E,
    E, e, e, e, e, e, e, e, e, E,
    E, e, e, e, e, e, e, e, e, E,
    E, E, E, E, E, E, E, E, E, E] )



###      Main Program     ###
def main(argv) :

    # Initialize Forest session and display the FOREST API version
    #       - The optional LIBRARIE boolean specifies whether or not the forest.lib opening and end-game book should be loaded, if present.
    #         That positions book dramatically improves evaluation accuracy and speed for game trees already known from FOREST
    # Note: the initialization might fail and raise an exception if:
    #       - The optional forestpy.dat file is not found or is corrupted
    #       - There is not enough RAM to allocate all what Forest needs (neural networks tensors, transposition tables, etc.)
    #       - The CPU doesn't support the AVX2 instruction set
    #       - The CPU has fewer than 4 cores (physical or virtual)
    #       The Forest session can be used for several evaluations, until a FOREST_TERMINER() ends the session
    try:
        VERSION = FOREST_INIT(LIBRAIRIE=True)
    except ForestError as ERREUR:
        print(ERREUR.args[0])
        exit()
    print("FOREST API version", VERSION)

    print()
    print("Evaluate Game 1 for X")
    # F5D6C3D3C4F4F6F3E6E7C6G6F8C5B4F7
    # Note: Games are in the WTHOR format for historical reasons
    GAME_1 = bytes([
    E, E, E, E, E, E, E, E, E, E,
    E, e, e, e, e, e, e, e, e, E,
    E, e, e, e, e, e, e, e, e, E,
    E, e, e, X, O, e, O, e, e, E,
    E, e, X, X, X, O, O, e, e, E,
    E, e, e, X, O, O, O, e, e, E,
    E, e, e, X, X, O, O, O, e, E,
    E, e, e, e, e, X, O, e, e, E,
    E, e, e, e, e, e, X, e, e, E,
    E, E, E, E, E, E, E, E, E, E] )

    # Display a Position et the console, in text format
    AFFICHER_JEU(GAME_1)

    # FOREST_NEURONE(GAME, PLAYER) returns the Value of the Position by the Value Neural Network of Forest
    # Note: this Value results from the static analysis of the position, without any search depth (i.e. Depth=0)
    VALUE = FOREST_NEURONE(GAME_1, BLACK)
    print("Neural Network Value =", VALUE)

    # FOREST_EVALUER() evaluates (estimates) a Position with Forest using an Alpha-Beta search
    # Param 1: the Position to be evaluated as a Byte vector in WTHOR format, White=1; Black=2, Empty=0, Edge=4
    # Param 2: the Player on the Position
    # Param 3: the search Depth (alpha-beta algorithm), between 2 and 14
    # Param 4: a boolean denoting whether all moves must be evaluated (True) or only the best ones (False)
    # Returns a list of tuples with each evaluated move and its relative value (a float) in the range [-64.0,+64.0]
    # Notes:
    #   - An Alpha-Beta search depth with Forest's value neural network as the evaluation function at the leaves
    #     dramatically improves the accurracy of evaluations. My interpretation is that the multiplication of the
    #     evaluations resulting from the tree search acts as a kind of "bagging" or "ensemble" of evaluations, hence
    #     reducing the average evaluation error
    #   - a max search depth of 14 may seem small compared to traditional programs that can reach 24 or more,
    #     but at the leafs of the search tree Forest uses a neural network which is much more accurate than
    #     those traditional programs, which compensates for the lower search depth
    DEPTH = 8
    print("Search depth", DEPTH)
    MOVES_EVAL = FOREST_EVALUER(GAME_1, BLACK, DEPTH, True)

    # Print all evaluated Moves
    for MOVE, VALUE in MOVES_EVAL:
        # REPRESENTATION(move) is a helper function giving the two character Ascii string representation of a move
        print("Move, Value:", REPRESENTATION(MOVE), "=", round(VALUE, 2))

    print()
    print("Solve (exactly) Game 2 for X")
    # 24 empty squares, Black plays D2-> Draw
    GAME_2 = bytearray(INIT_GAME)

    # A Key uniquely identifies a Position, regardless of the 8 possible symetries of Othello
    # Note: the two integers of a key are NOT the bitboards of a position. This is because SQLite does not support 64 bit unsigned integers
    KEY = (90624953756491, 2546047750800)
    # Fills a position according to a Key
    CALCUL_POSITION(GAME_2, KEY)
    AFFICHER_JEU(GAME_2)
    # Compute the Unique Key of a Position
    print("Key =", CLE_UNIQUE(GAME_2))

    # Display the Neural Network immediate Policy as an example to be compared to the exact solution
    print("Neural Network Policy =", [REPRESENTATION(MOVE) for MOVE in FOREST_CLASSER(GAME_2, BLACK)])

    # Display the Neural Network immediate Value as an example to be compared to the exact solution
    print("Neural Network Value  =", FOREST_NEURONE(GAME_2, BLACK))

    # FOREST_RESOUDRE() solves (exactly) a Position with Forest
    # Param 1: the Position to be evaluated as a Byte vector in WTHOR format, White=1; Black=2, Empty=0, Edge=4
    # Param 2: the Player on the Position
    # Param 3: a boolean denoting whether all moves must be evaluated (True) or only the best ones (False)
    # Returns a list of tuples with each evaluated move and its relative score (an integer) in the range [-64,+64]
    # Notes:
    #   - The search depth is the number of empty squares on the Position
    #   - Forest is too slow for solving positions with more than 25 empty squares, EDAX should be used for higher numbers
    MOVES_SOLVED = FOREST_RESOUDRE(GAME_2, BLACK, False)

    # Print all evaluated Moves
    for MOVE, VALUE in MOVES_SOLVED:
        print("Best Move, Exact Value:", REPRESENTATION(MOVE), "=", VALUE)

    print()
    print("Search Win/Loss/Draw in Game 2 for X")

    # FOREST_WLD() determines the WLD pronostic of a Position with Forest
    #              i.e. whether the position is Win, Loss, Draw for the player
    # Param 1: the Position to be evaluated as a Byte vector in WTHOR format, White=1; Black=2, Empty=0, Edge=4
    # Param 2: the Player on the Position
    # Returns a list of tuples with each evaluated move its pronostic (an integer): -1 for a loss, 0 for a draw, +1 for a win
    # Notes:
    #   - The search depth is the number of empty squares on the Position
    #   - The best move is not necessarily the first in the list
    #   - Determining the WLD pronostic is fater than exactly solving the position
    #   - Not all the moves are evaluated, because moves that are too bad to have a chance to Win/Draw are neglected
    MOVES_WLD = FOREST_WLD(GAME_2, BLACK)

    # Print all evaluated Moves
    for MOVE, PRONOSTIC in MOVES_WLD:
        if PRONOSTIC == +1:
            WLD = "Win"
        elif PRONOSTIC == 0:
            WLD = "Draw"
        elif PRONOSTIC == -1:
            WLD = "Loss"
        else:
            WLD = "unknown"
        print("Move, Pronostic:", REPRESENTATION(MOVE), "->", WLD)

    # Terminate Forest session
    FOREST_TERMINER()

    # Other useful helper functions in the OTHPY Python extension

    # NOMBRE_PIONS(GAME, int PLAYER) returns the number of Player Discs on the Game
    print()
    AFFICHER_JEU(GAME_2)
    print("There are", NOMBRE_PIONS(GAME_2, BLACK), "X discs on the game")
    print("There are", NOMBRE_PIONS(GAME_2, WHITE), "O discs on the game")

    # LISTE_COUPS(str STRING, int NUMBER) returns the list of NUMBER Moves in the Ascii string representation of a series of Moves
    print()
    MOVE_STRING = "F5D6C3D3C4F4F6F3E6E7C6G6F8C5B4F7"
    MOVE_LIST = LISTE_COUPS(MOVE_STRING, 16)
    print("Moves in", MOVE_STRING, "=", MOVE_LIST)

    # CHAINE_PARTIE(list MOVES) is the reverse function
    print()
    print("Ascii representation of", MOVE_LIST, "=", CHAINE_PARTIE(MOVE_LIST))

    # JOUER_LISTE_COUPS(GAME, list MOVES, int PLAYER) plays a series of MOVES on the GAME starting with PLAYER
    # returns the final GAME, the NUMBER of moves actually playerd, and the next PLAYER
    print()
    print("Play moves", MOVE_STRING, "from start position")
    GAME_3 = bytearray(INIT_GAME)
    N, PLAYER = JOUER_LISTE_COUPS(GAME_3, MOVE_LIST, BLACK)
    AFFICHER_JEU(GAME_3)
    print(N, "moves played")
    print("Next player", "X" if PLAYER==BLACK else "O")


    # LISTE_JOUABLES(GAME, int PLAYER) returns the list of the licit Moves on the Game
    print()
    print("Licit moves on Game")
    AFFICHER_JEU(GAME_2)
    print(LISTE_JOUABLES(GAME_2, BLACK))

    # INVERSER_PIONS(GAME) permutes Black and White discs on the Game
    print()
    AFFICHER_JEU(GAME_2)
    print("Permuted:")
    GAME_3 = bytearray(GAME_2)
    INVERSER_PIONS(GAME_3)
    AFFICHER_JEU(GAME_3)

    # JOUABLE(GAME, int MOVE, int PLAYER) tells whether a Move on the Game is licit for a Player
    print()
    AFFICHER_JEU(GAME_2)
    MOVE = 11
    if JOUABLE(GAME_2, MOVE, BLACK):
        print(REPRESENTATION(MOVE), "is licit for X")
    else:
        print(REPRESENTATION(MOVE), "is illicit for X")
    MOVE = 51
    if JOUABLE(GAME_2, MOVE, BLACK):
        print(REPRESENTATION(MOVE), "is licit for X")
    else:
        print(REPRESENTATION(MOVE), "is illicit for X")

    # JOUER_COUP(GAME, int MOVE, int PLAYER) plays a Move for Player on the Game, returns True if the move was licit, False otherwise
    print()
    MOVE = 51
    print("Play move", REPRESENTATION(MOVE), "on:")
    AFFICHER_JEU(GAME_2)
    GAME_3 = bytearray(GAME_2)
    if JOUER_COUP(GAME_3, MOVE, BLACK):
        print("Valid move:")
        AFFICHER_JEU(GAME_3)

    # COINCE(GAME, int PLAYER) returns True if the Player is blocked (must pass), False if Player still has at list a valid Move
    print()
    AFFICHER_JEU(GAME_2)
    if not COINCE(GAME_2, BLACK):
        print("X can play")
    else:
        print("X must pass")

    # NOMBRE_JOUABLES(GAME, int PLAYER) returns the number of Licit Moves for Player on the Game
    print()
    AFFICHER_JEU(GAME_2)
    print("X can play", NOMBRE_JOUABLES(GAME_2, BLACK), "moves")

    # DEFINITIFS(GAME) returns estimates of the numbers of Stable Black and White Disks
    # Note: the numbers are possibly under-estimated, because some complex patterns of stable discs are difficult to identify
    print()
    GAME_3 = bytes([
    E, E, E, E, E, E, E, E, E, E,
    E, X, X, X, X, X, O, O, O, E,
    E, O, X, X, O, O, O, O, O, E,
    E, e, e, X, O, e, O, O, O, E,
    E, e, X, X, X, O, O, e, e, E,
    E, e, e, X, O, O, O, e, e, E,
    E, e, e, X, X, O, O, O, X, E,
    E, e, e, e, e, X, O, O, X, E,
    E, e, e, e, e, e, X, X, X, E,
    E, E, E, E, E, E, E, E, E, E] )
    AFFICHER_JEU(GAME_3)
    print("Stable disks for Black and White =", DEFINITIFS(GAME_3))
    
    # CALCUL_BITBOARDS(GAME) returns the Game Bitboards for Black and White
    print()
    AFFICHER_JEU(GAME_2)
    print("Bitboards for Black and White", CALCUL_BITBOARDS(GAME_2))

    print("End.")

if __name__ == "__main__":
    main(sys.argv)           
